;; Weather Station
;;
;; Copyright 2020 - By Michael Kohn
;; http://www.mikekohn.net/
;; mike@mikekohn.net
;;
;; Circuit for a weather station that SparkFun sells. Has wind speed,
;; wind direction, and rain fall. Adding a DS18B20 for temperature.
;;
;; https://www.sparkfun.com/products/15901

;; Wind direction runs 3.3v with 10k ohm resistor on top.
;;   DIR    OHMS      ADC  ADC  COMPASS
;;   0.0   33000   0x0311 785   N
;;  22.5    6570   0x0195 405
;;  45.0    8200   0x01cc 460   NE
;;  67.5     891   0x0053  83
;;  90.0    1000   0x005d  93   E
;; 112.5     688   0x0041  65
;; 135.0    2200   0x00b8 184   SE
;; 157.5    1410   0x007e 126
;; 180.0    3900   0x011f 287   S
;; 202.5    3140   0x00f4 244
;; 225.0   16000   0x0275 629   SW
;; 247.5   14120   0x0256 598
;; 270.0  120000   0x03b0 944   W
;; 292.5   42120   0x033a 826
;; 315.0   64900   0x0376 886   NW
;; 337.5   21880   0x02be 702

;; Wind speed count is 2.4km/h per count.
;; Rain count is 0.2794mm of rain.

.include "msp430x2xx.inc"

RAM equ 0x0200
WIND_SPEED_COUNT equ RAM
WIND_DIRECTION_VALUE equ RAM+2
RAIN_COUNT equ RAM+4

DS18B20_DATA equ 0x0210
DS18B20_CMD equ 0x0220
DS18B20_CMD_LEN equ 0x0221
DS18B20_BIT equ 0x0222
DEBUG_STATE equ 0x0223

TEMPERATURE equ DS18B20_DATA

;; Port 1
WIND_SPEED equ 0x01
WIND_DIRECTION equ 0x20

;; Port 2
DS18B20 equ 0x08
RAIN equ 0x10

;  r4 =
;  r5 =
;  r6 =
;  r7 =
;  r8 =
;  r9 = count interrupt_b
; r10 =
; r11 = count in interrupt_a
; r12 = temp in interrupt_b
; r13 =
; r14 =
; r15 = temp

.org 0xc000
start:
  ;; Turn off watchdog
  mov.w #WDTPW|WDTHOLD, &WDTCTL

  ;; Turn off interrupts
  dint

  ;; Setup stack pointer
  mov.w #0x0400, SP

  ;; Set MCLK to 2 MHz with DCO
  mov.b #DCO_3, &DCOCTL
  mov.b #RSEL_9, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Set SMCLK to 32.768kHz external crystal
  mov.b #XCAP_3, &BCSCTL3

  ;; Setup output pins
  ;; P1.0 = Wind speed
  ;; P1.1 = RX
  ;; P1.2 = TX
  ;; P1.5 = Wind direction (analog)
  ;; P2.3 = DS18B20
  ;; P2.4 = Rain
  ;; P2.6 = 32.768Hz Crystal
  ;; P2.7 = 32.768Hz Crystal
  mov.b #0, &P1DIR
  mov.b #0, &P1OUT
  mov.b #0x07, &P1SEL
  mov.b #0x06, &P1SEL2
  mov.b #DS18B20, &P2DIR
  mov.b #DS18B20, &P2OUT

  ;; Setup A/D
  mov.w #ADC10SHT_3|ADC10ON, &ADC10CTL0
  mov.w #INCH_5|ADC10SSEL_3, &ADC10CTL1
  mov.b #WIND_DIRECTION, &ADC10AE0

  ;; Setup UART (9600 @ 2 MHz)
  mov.b #UCSSEL_2|UCSWRST, &UCA0CTL1
  mov.b #0, &UCA0CTL0
  mov.b #UCBRS_2, &UCA0MCTL
  mov.b #220, &UCA0BR0
  mov.b #0, &UCA0BR1
  bic.b #UCSWRST, &UCA0CTL1

  ;; Setup Timer B
  mov.w #32768, &TBCCR0        ; 32768 ticks = 1 second
  mov.w #TBSSEL_1|MC_1, &TBCTL ; ACLK, DIV1, COUNT to TBCCR0
  mov.w #CCIE, &TBCCTL0
  mov.w #0, &TBCCTL1

  mov.w #0, &WIND_SPEED_COUNT
  mov.w #0, &WIND_DIRECTION_VALUE
  mov.w #0, &RAIN_COUNT
  mov.w #0, &TEMPERATURE

  ;; Turn on interrupts
  eint

  ;; To help debugging, clear out some RAM.
  mov.w #0x200, r15
memset:
  mov.w #0, 0(r15)
  add.w #2, r15
  cmp.w #0x230, r15
  jnz memset

  mov.w #0, r9
  mov.w #0, r6

  call #set_timer_a_for_ds18b20
  call #read_temp
  call #set_timer_a_for_windspeed

main:
  mov.w &P2IN, r15
  bic.w #0xef, r15
  cmp.w r15, r6
  jz main_no_rain_change
  inc.w &RAIN_COUNT
  mov.w r15, r6
main_no_rain_change:
  cmp.w #60, r9
  jlo main

  clrc
  rra &RAIN_COUNT
  call #send_json
  call #set_timer_a_for_ds18b20
  call #read_temp
  call #set_timer_a_for_windspeed
  mov.w #0, r9
  ;mov.w #0, &WIND_SPEED_COUNT
  mov.w #0, &RAIN_COUNT
  jmp main

send_json:
  mov.w #json_wind_direction, r14
  call #uart_send_string
  mov.w #WIND_DIRECTION_VALUE, r14
  call #uart_send_value

  mov.w #json_wind_speed, r14
  call #uart_send_string
  mov.w #WIND_SPEED_COUNT, r14
  call #uart_send_value

  mov.w #json_rain, r14
  call #uart_send_string
  mov.w #RAIN_COUNT, r14
  call #uart_send_value

  mov.w #json_temperature, r14
  call #uart_send_string
  mov.w #TEMPERATURE, r14
  call #uart_send_value

  mov.w #json_end, r14
  call #uart_send_string
  ret

; uart_send_char(r15)
uart_send_char:
  bit.b #UCA0TXIFG, &IFG2
  jz uart_send_char
  mov.b r15, &UCA0TXBUF
  ret

uart_send_string:
  mov.b @r14+, r15
  cmp.b #0, r15
  jz uart_send_string_exit
  call #uart_send_char
  jmp uart_send_string
uart_send_string_exit:
  ret

uart_send_hex_nybble:
  cmp.b #0xa, r15
  jl uart_send_hex_nybble_09
  add.b #'a'-10, r15
  call #uart_send_char
  ret
uart_send_hex_nybble_09:
  add.b #'0', r15
  call #uart_send_char
  ret

uart_send_byte:
  mov.b @r14, r15
  rra.w r15
  rra.w r15
  rra.w r15
  rra.w r15
  call #uart_send_hex_nybble
  mov.b @r14, r15
  bic.b #0xf0, r15
  call #uart_send_hex_nybble
  sub.w #1, r14
  ret

uart_send_value:
  add.w #1, r14
  call #uart_send_byte
  call #uart_send_byte
  ret

timer_interrupt_a:
  inc.w r11
  reti

timer_interrupt_b:
  add.w #1, r9

  mov.w &TAR, r12
  cmp.w &WIND_SPEED_COUNT, r12
  jlo timer_interrupt_b_not_max_speed
  mov.w r12, &WIND_SPEED_COUNT
timer_interrupt_b_not_max_speed:
  mov.w #0, &TAR

  ;; Read 10 bit DAC and shift to 12 bit.
  mov.w &ADC10MEM, r12
  mov.w r12, &WIND_DIRECTION_VALUE

  ;; Start new ADC encoding.
  bis.w #ENC|ADC10SC, &ADC10CTL0
  reti

off:
  bic.b #8, &P2OUT
  reti

set_timer_a_for_windspeed:
  ;; Connected to wind-speed.
  dint
  ;mov.w #0xffff, &TACCR0
  mov.w #TASSEL_0|MC_2, &TACTL ; TACLK, DIV1, COUNT to 0xffff
  mov.w #0, &TACCTL0           ; No interrupt.
  mov.w #0, &TACCTL1
  mov.w #0, &TAR
  mov.w #0, &WIND_SPEED_COUNT
  eint
  ret

set_timer_a_for_ds18b20:
  ;; Needs to be a timer.
  mov.w #250, &TACCR0
  mov.w #TASSEL_2|MC_1, &TACTL ; SMCLK, DIV1, COUNT to TACCR0
  mov.w #CCIE, &TACCTL0
  mov.w #0, &TACCTL1
  ret

.include "ds18b20.inc"

json_wind_direction:
  .asciiz "{\r\n  \"wind_direction\": \"0x"
json_wind_speed:
  .asciiz "\",\r\n  \"wind_speed\": \"0x"
json_rain:
  .asciiz "\",\r\n  \"rain\": \"0x"
json_temperature:
  .asciiz "\",\r\n  \"temperature\": \"0x"
json_end:
  .asciiz "\"\r\n}\r\n"

.org 0xfff2
  dw timer_interrupt_a     ; Timer0_A3 TACCR0, CCIFG
.org 0xfffa
  dw timer_interrupt_b     ; Timer1_A3 TACCR0, CCIFG
.org 0xfffe
  dw start                 ; Reset

